#!/usr/bin/env python

"""
A simple example of parallel computation using message exchanges directly,
exchanges and managed callables, and map-style processing with parallel-enabled
functions.
"""

import pprocess
import time

# Array size and a limit on the number of processes.

N = 10
limit = 10
delay = 1

def make_array(n):

    "Make an 'n' * 'n' array initialised with zeros."

    return [list(x) for x in [(0,) * n] * n]

# Work function and monitoring class.

def calculate(ch, i, j):

    """
    A time-consuming calculation, using 'ch' to communicate with the parent
    process, with 'i' and 'j' as operands.
    """

    time.sleep(delay)
    ch.send((i, j, i * N + j))

class MyExchange(pprocess.Exchange):

    "Parallel convenience class containing the array assignment operation."

    def store_data(self, ch):
        i, j, result = ch.receive()
        self.D[i][j] = result

# Alternative work function.

def calculate_seq(i, j):

    """
    A version of 'calculate' with a simpler result suitable for sequential
    collection. It accepts no channel, pretending to be a function used in a
    non-parallel context.
    """

    time.sleep(delay)
    return i * N + j

# The alternative work function converted into a parallel-enabled version.

calculate2 = pprocess.MakeParallel(calculate_seq)

# Main program.

if __name__ == "__main__":

    # Initialise the communications exchange with a limit on the number of
    # channels/processes.

    exchange = MyExchange(limit=limit)

    # Initialise an array - it is stored in the exchange to permit automatic
    # assignment of values as the data arrives.

    exchange.D = make_array(N)

    # The parallel computation.

    print "Calculating..."
    for i in range(0, N):
        for j in range(0, N):
            exchange.start(calculate, i, j)
            print i, j

    print "Finishing..."
    exchange.finish()

    # Show the result.

    print
    for row in exchange.D:
        print row

    # Try again with managed callables.

    exchange.D = make_array(N)
    calc = exchange.manage(calculate)

    print "Calculating..."
    for i in range(0, N):
        for j in range(0, N):
            calc(i, j)
            print i, j

    print "Finishing..."
    exchange.finish()

    print
    for row in exchange.D:
        print row

    # Try again with a map, building the results sequentially.

    mymap = pprocess.pmap
    results = []

    print "Calculating..."
    for i in range(0, N):
        sequence = []
        for j in range(0, N):
            sequence.append((i, j))
        results.append(mymap(calculate2, sequence))
        print i

    print
    for row in results:
        print row

# vim: tabstop=4 expandtab shiftwidth=4
